# -*- coding: utf-8 -*-

import logging

from openerp import api, fields, models
from openerp.exceptions import ValidationError
from openerp.tools.translate import _


_logger = logging.getLogger(__name__)


def _sanitize(text):
    # Chars not allowed on CFDI
    chars = "\\`*_{}[]()>#+-.!$/"  # noqa
    # Replace all characters based on best performance answer as found on:
    # https://stackoverflow.com/a/27086669/2817675
    for c in chars:
        if c in text:
            text = text.replace(c, " ")
    return text


class BaseCfdi(models.AbstractModel):
    """Include this model within models inherited by an object to include
    basic fields, properties and functions needed for create a CFDI from any
    object.
    """

    _name = "base.cfdi"
    _description = __doc__

    @property
    def xml_name(self):
        fname = ""
        if self.state != "draft":
            lista = [self.company_id.partner_id.vat_split]
            if hasattr(self, "number") and self.number:
                lista.append(self.number)
            elif hasattr(self, "internal_number") and self.internal_number:
                lista.append(self.internal_number)

            fname = "_".join(lista)
        return fname

    @property
    def serie(self):
        """Return serie for display on CFDI"""
        self.ensure_one()
        # Get interpolated sequence
        if self.journal_id.sequence_id.prefix:
            d = self.env["ir.sequence"]._interpolation_dict_context()
            serie = self.journal_id.sequence_id._interpolate(
                self.journal_id.sequence_id.prefix, d
            )
        else:
            serie = ""
        return _sanitize(serie).strip()

    @property
    def folio(self):
        """Return folio for display on CFDI"""
        self.ensure_one()
        folio = _sanitize(self.number or self.internal_number)
        return folio.replace(self.serie, "").strip()

    cfdi_id = fields.Many2one(
        "ir.attachment.facturae.mx",
        "CFDI",
        copy=False,
        help="CFDI realed to the selected record.",
    )
    cfdi_datetime = fields.Datetime()
    cfdi_folio_fiscal = fields.Char(
        "CFD-I Folio Fiscal",
        related="cfdi_id.uuid",
        help="Folio used in the electronic invoice",
    )
    cfdi_state = fields.Selection(related="cfdi_id.state")
    cfdi_relation_type = fields.Many2one(
        "cfdi.relation.type", "CFDI Relation type", copy=False
    )
    related_cfdi_ids = fields.Many2many(
        "ir.attachment.facturae.mx",
        "cfdis_related_rel",
        "related_cfdi_id",
        "original_cfdi_id",
        "Refund invoices",
        compute="_compute_related_cfdi_ids",
        readonly=True,
        copy=False,
        help="Original CFDI to which this CFDI is referred to",
    )
    cfdi_adenda_id = fields.Many2one(
        "cfdi.adenda",
        "Addendum",
        readonly=True,
        states={"draft": [("readonly", False)]},
        help="Select addendum node to use on this CFDI."
    )

    @api.multi
    def _compute_related_cfdi_ids(self):
        for record in self:
            cfdi_related = self.env["cfdi.related"].search(
                [
                    ("related_cfdi_id", "=", record.id),
                    ("type_attachment", "=", record._name),
                ]
            )
            record.update({"related_cfdi_ids": cfdi_related.original_cfdi_id})

    @api.one
    def create_cfdi(self):
        """Creates a new cfdi object related with current record"""
        self.cfdi_datetime = self._get_cfdi_datetime()
        self.cfdi_id = self.env["ir.attachment.facturae.mx"].create(
            {
                "name": self.xml_name,
                "res_id": self.id,
                "type_attachment": self._name,
                "company_id": self.company_id.id,
                "cfdi_adenda_id": self.cfdi_adenda_id.id or None,
            }
        )
        self.cfdi_id.action_validate()

    @api.multi
    def _get_cfdi_datetime(self):
        """Function that allow define several ways to get cfdi_date
        according to the document we are working on.
        """
        self.ensure_one()
        return fields.Datetime.now()

    @api.one
    def action_validate_cfdi(self):
        """Validates a CFDI related with current record"""
        # First try to validate the file as is, this way we ask PAC to give us the
        # uuid assigned for given XML in case there were a communication issue on
        # our side.
        self.cfdi_id.action_validate()
        # In case cfdi is not signed yet try recreating xml
        if self.cfdi_id.state not in ("signed", "done"):
            self.cfdi_datetime = fields.Datetime.now()
            self.cfdi_id.with_context(recreate_xml=True).action_validate()

    @api.one
    def cancel_cfdi(self):
        """ Cancels the cfdi related with current record
        return True if cancel proceed or false if must wait
        """
        if self.cfdi_id and self.cfdi_id.state in ["signed", "done", "waiting"]:
            return self.cfdi_id.action_cancel()

    def consult_cfdi_cancellation_status(self):
        """Consult the status cancel of the cfdi related"""
        if self.cfdi_id and self.cfdi_id.state in ["waiting", "cancel"]:
            return self.cfdi_id.action_consult_cancellation_status()

    @api.multi
    def _cancel_cfdi_related_document(self):
        raise ValidationError(_("Cfdi was not canceled due to an internal problem"))

    @api.multi
    def _cancel_cfdi(self):
        """General steps needed for cancel a CFDI related to a record"""
        # self._origin_cfdi_search()
        self.ensure_one()
        # Before cancel original voucher set relation to current related CFDI
        self.write(
            {"cfdi_relation_type": self.env.ref("l10n_mx.cfdi_relation_type_04").id}
        )
        self.env["cfdi.related"].search(
            [("related_cfdi_id", "=", self.id), ("type_attachment", "=", self._name)]
        ).unlink()
        self.env["cfdi.related"].create(
            {
                "related_cfdi_id": self.id,
                "original_cfdi_id": self.cfdi_id.id,
                "type_attachment": self._name,
            }
        )
        # We also remove relation for attachments to avoid confusing situations
        self.cfdi_id.file_xml_sign.res_id = None
        if self.cfdi_id.file_pdf:
            self.cfdi_id.file_pdf.res_id = None
        # Actually cancel de related document
        self._cancel_cfdi_related_document()
