# -*- coding: utf-8 -*-

import base64
import hashlib
import logging
import pkg_resources
import re
import six
from dateutil.parser import parse
from lxml import etree as ET  # noqa
from pytz import timezone, utc
from requests import Request, Session
from requests.exceptions import HTTPError
from xml.dom import minidom
from xml.sax.saxutils import escape

import M2Crypto as m2  # noqa

from openerp import api, fields, models
from openerp.exceptions import ValidationError, Warning as UserError
from openerp.report import render_report
from openerp.tools.translate import _


# Namespaces to parse CFDI files
NAMESPACES = {
    "cfdi": "http://www.sat.gob.mx/cfd/3",
    "tfd": "http://www.sat.gob.mx/TimbreFiscalDigital",
}

# Field names to parse or validate XML according to version
NAMES = {
    "3.2": {
        "xslt_file": "cadenaoriginal_3_2.xslt",
        "hash_type": "sha1",
        "stamp": "sello",
        "stamp_sat": "selloSAT",
        "cert": "noCertificado",
        "cert_sat": "noCertificadoSAT",
        "tid": "rfc",
        "date": "fecha",
        "total": "total",
    },
    "3.3": {
        "xslt_file": "cadenaoriginal_3_3.xslt",
        "hash_type": "sha256",
        "stamp": "Sello",
        "stamp_sat": "SelloSAT",
        "cert": "NoCertificado",
        "cert_sat": "NoCertificadoSAT",
        "tid": "Rfc",
        "date": "Fecha",
        "total": "Total",
    },
}

# Definition for SAT SOAP service for validate UUIDs
SOAPENV_SAT_VALIDATE = """<?xml version="1.0" encoding="UTF-8"?>
<soapenv:Envelope
    xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/"
    xmlns:tem="http://tempuri.org/">
    <soapenv:Header/>
    <soapenv:Body>
        <tem:Consulta>
            <tem:expresionImpresa>
                ?re={re}&amp;rr={rr}&amp;tt={tt}&amp;id={id}
            </tem:expresionImpresa>
        </tem:Consulta>
    </soapenv:Body>
</soapenv:Envelope>
"""

_logger = logging.getLogger(__name__)


def parse_xml_string(xmlstr):
    """Parse an XML string (binary or unicode) with a default parser.
    Function almost copy/paste from libtaxii developed by @gtback
    Thank you!!!
    :param xmlstr: An XML String to parse
    :return: an etree._Element
    """
    if isinstance(xmlstr, six.binary_type):
        xmlstr = six.BytesIO(xmlstr)
    elif isinstance(xmlstr, six.text_type):
        # LXML doesn't accept Unicode strings with an explicit encoding, so
        # try to detect and encode to bytes before passing to LXML.
        encoding = re.findall(r'encoding=[\'"]([0-9A-Za-z_\-]+)', xmlstr[:50], re.I)
        # re.findall returns a list of matching strings. We only care about the
        # first one.
        if encoding:
            xmlstr = six.BytesIO(xmlstr.encode(encoding[0]))
        else:
            # We assume that all xmlstr that arrives here are unicode and with
            # UTF-8 codification, this allow us to parse XML files without an
            # explicit XML declaration line
            xmlstr = six.BytesIO(xmlstr.encode("UTF-8"))

    # Construct a ``etree.XMLParser`` with ``no_network=True`` and
    # ``huge_tree=False``.
    parser = ET.XMLParser(
        attribute_defaults=False,
        dtd_validation=False,
        load_dtd=False,
        no_network=True,
        ns_clean=True,
        recover=True,
        remove_blank_text=False,
        remove_comments=False,
        remove_pis=False,
        strip_cdata=True,
        compact=True,
        resolve_entities=False,
        huge_tree=False,
    )

    return ET.parse(xmlstr, parser).getroot()


class IrAttachmentFacturaeMx(models.Model):
    # pylint: disable=R0904
    _name = "ir.attachment.facturae.mx"
    _inherit = ["mail.thread", "ir.needaction_mixin"]
    _rec_name = "uuid"
    _order = "fecha DESC"

    _states = [
        ("draft", _("Draft")),
        ("signed", _("Signed")),
        ("done", _("Done")),
        ("waiting", _("To cancel")),
        ("cancel", _("Cancelled")),
    ]

    @api.model
    def _get_default_company(self):
        company_id = self.env.user._get_company()
        if not company_id:
            raise UserError(_("There is no default company for the current user!"))
        return self.env["res.company"].browse(company_id)

    name = fields.Char(
        size=128, required=True, readonly=True, help="Name of attachment generated"
    )
    res_id = fields.Integer(
        "Resource id", readonly=True, help="Invoice to which it belongs this attachment"
    )
    file_xml = fields.Many2one(
        "ir.attachment", readonly=True, help="XML file before PAC signature",
    )
    file_xml_sign = fields.Many2one(
        "ir.attachment", "File XML Sign", readonly=True, help="File XML signed"
    )
    file_pdf = fields.Many2one(
        "ir.attachment",
        "File PDF",
        readonly=True,
        help="Report PDF generated for the electronic Invoice",
    )
    cfdi_type = fields.Selection(
        [("incoming", "Incoming"), ("outgoing", "Outgoing")],
        "Move Type",
        compute="_compute_cfdi_type",
        store=True,
        readonly=True,
        help="Type of Electronic Invoice",
    )
    state = fields.Selection(
        _states,
        readonly=True,
        default="draft",
        required=True,
        help="State of attachments",
    )
    type_attachment = fields.Selection(
        selection=[
            ("account.invoice", _("Invoice")),
            ("account.voucher", _("Payment")),
            ("hr.payslip", _("Payslip")),
            ("stock.picking", _("Picking")),
        ],
        string="Document Type",
        readonly=True,
    )
    config_id = fields.Many2one(
        "ir.attachment.facturae.mx.config",
        compute="_compute_config_id",
        store=True,
        help="Configuration used for validate this CFDI agains SAT.",
    )
    sello = fields.Text(
        "Stamp", size=512, copy=False, compute="_compute_sello", help="Digital Stamp"
    )
    cfdi_sello = fields.Text(
        "CFD-I Stamp",
        copy=False,
        compute="_compute_cfdi_sello",
        help="Sign assigned by the SAT",
    )
    cfdi_fecha_timbrado = fields.Datetime(
        "CFD-I Date Stamping",
        copy=False,
        store=True,
        compute="_compute_cfdi_fecha_timbrado",
        help="Date when is stamped the electronic invoice",
    )
    cfdi_fecha_cancelado = fields.Datetime(
        "CFD-I Date Cancel",
        copy=False,
        store=True,
        compute="_compute_cfdi_fecha_timbrado",
        help="Date when is canceled the CFDI",
    )
    no_certificado = fields.Text(
        "No. Certificate",
        size=64,
        copy=False,
        compute="_compute_no_certificado",
        help="Number of serie of certificate used for the invoice",
    )
    cfdi_no_certificado = fields.Char(
        "CFD-I Certificado",
        size=32,
        copy=False,
        compute="_compute_cfdi_no_certificado",
        help="Serial Number of the Certificate",
    )
    company_id = fields.Many2one(
        "res.company",
        string="Company",
        required=True,
        readonly=True,
        states={"draft": [("readonly", False)]},
        default=_get_default_company,
        help="Company to which it belongs this attachment",
    )
    uuid = fields.Char(compute="_compute_uuid", store=True, track_visibility="onchange")
    cfdi_cadena_original = fields.Text(
        "CFD-I Original String",
        copy=False,
        compute="_compute_cfdi_cadena_original",
        help="Original String used in the electronic invoice",
    )
    version = fields.Char(compute="_compute_version")
    fecha = fields.Datetime(compute="_compute_fecha", store=True, copy=False)
    rfc_emisor = fields.Char(compute="_compute_vat_numbers", store=True)
    rfc_receptor = fields.Char(compute="_compute_vat_numbers")
    total = fields.Float(compute="_compute_total")
    version_timbre = fields.Char(compute="_compute_version_timbre")
    cfdi_adenda_id = fields.Many2one(
        "cfdi.adenda",
        "Addendum",
        readonly=True,
        states={"draft": [("readonly", False)]},
        help="Addenda used for this cfdi."
    )

    @api.one
    @api.depends("file_xml_sign", "rfc_emisor", "company_id")
    def _compute_cfdi_type(self):
        # Default to cfdi_type already set for record or incoming as fallback
        cfdi_type = self.cfdi_type if self.cfdi_type else "incoming"
        # Try to get cfdi_type from XML file
        if self.file_xml_sign and not self.cfdi_type:
            cfdi_type = (
                "incoming"
                if self.rfc_emisor == self.company_id.partner_id.vat_split
                else "outgoing"
            )

        self.cfdi_type = cfdi_type

    @api.one
    @api.depends("cfdi_type", "type_attachment")
    def _compute_config_id(self):
        """Computes the corresponding rule for validate CFDI against SAT.
        The selected rule will be used for sign SAT when signing the CFDI
        """
        ir_attachment_config_obj = self.env["ir.attachment.facturae.mx.config"]
        # Search for a template which fills model requirements
        domain = [("model", "=", self.type_attachment)]
        config = ir_attachment_config_obj.search(domain, limit=1)

        if self.cfdi_type == "incoming" and not config:
            raise UserError(
                _("You do not have any defined template " "for {model} model").format(
                    model=self.type_attachment
                )
            )
        elif self.cfdi_type == "incoming":
            self.config_id = config.id
        else:
            self.config_id = False

    @api.one
    @api.depends("file_xml_sign", "version")
    def _compute_sello(self):
        """Get sello field from XML file"""
        self.sello = self._parse_xml(NAMES[self.version]["stamp"])

    @api.one
    @api.depends("file_xml_sign", "version")
    def _compute_cfdi_sello(self):
        """Get selloSAT field from XML file"""
        self.cfdi_sello = self._parse_xml(NAMES[self.version]["stamp_sat"])

    @api.one
    @api.depends("file_xml_sign")
    def _compute_cfdi_fecha_timbrado(self):
        """Get FechaTimbrado field from XML file"""
        if self.file_xml_sign:
            self.cfdi_fecha_timbrado = self._parser_to_date_from_xml("FechaTimbrado")

    @api.one
    @api.depends("file_xml_sign", "version")
    def _compute_no_certificado(self):
        """Get noCertificado field from XML file"""
        self.no_certificado = self._parse_xml(NAMES[self.version]["cert"])

    @api.one
    @api.depends("file_xml_sign", "version")
    def _compute_cfdi_no_certificado(self):
        """Get noCertificadoSAT field from XML file"""
        self.cfdi_no_certificado = self._parse_xml(NAMES[self.version]["cert_sat"])

    @api.one
    @api.depends("file_xml_sign")
    def _compute_uuid(self):
        """Get uuid field from XML file"""
        uuid = self._parse_xml("UUID", element="tfd:TimbreFiscalDigital")
        if uuid:
            uuid = uuid.upper()
        self.uuid = uuid

    @api.one
    @api.depends(
        "version_timbre",
        "version",
        "uuid",
        "cfdi_fecha_timbrado",
        "sello",
        "no_certificado",
    )
    def _compute_cfdi_cadena_original(self):
        """Get _compute_cfdi_cadena_original field"""
        if self.file_xml_sign:
            cadena = ("|").join(
                [
                    self.version_timbre,
                    self.uuid,
                    self.cfdi_fecha_timbrado,
                    self.sello,
                    self.no_certificado,
                ]
            )
            self.cfdi_cadena_original = "||{0}||".format(cadena)

    @api.one
    @api.depends("config_id", "file_xml_sign")
    def _compute_version(self):
        try:
            version = self._parse_xml("Version", element="cfdi:Comprobante")
        except AttributeError:
            version = self._parse_xml("version", element="cfdi:Comprobante")
        self.version = version or self.config_id.version

    @api.one
    @api.depends("file_xml_sign")
    def _compute_version_timbre(self):
        try:
            version_timbre = self._parse_xml(
                "Version", element="tfd:TimbreFiscalDigital"
            )
        except AttributeError:
            version_timbre = self._parse_xml(
                "version", element="tfd:TimbreFiscalDigital"
            )
        self.version_timbre = version_timbre

    @api.one
    @api.depends("file_xml_sign", "version")
    def _compute_fecha(self):
        """Get issuing date from XML file"""
        if self.file_xml_sign:
            self.fecha = self._parser_to_date_from_xml(NAMES[self.version]["date"])

    @api.one
    @api.depends("file_xml_sign", "version")
    def _compute_vat_numbers(self):
        """Get issuing and billing companies VAT numbers from XML file"""
        self.rfc_emisor = self._parse_xml(
            NAMES[self.version]["tid"], element="cfdi:Emisor"
        )
        self.rfc_receptor = self._parse_xml(
            NAMES[self.version]["tid"], element="cfdi:Receptor"
        )

    @api.one
    @api.depends("file_xml_sign", "version")
    def _compute_total(self):
        """Get total from XML file"""
        total = self._parse_xml(
            NAMES[self.version]["total"], element="cfdi:Comprobante"
        )
        try:
            total = float(total)
        except ValueError:
            total = 0.0
        self.total = total

    @api.model
    def _parser_to_date_from_xml(self, tag):
        """Get a datime value from a value in a xml

        Get a datime value parsing an xml file looking the element with 'tag',  try to
        parse the value found if any, and if not throw an exception, catch it and
        return None

        :param tag: tag to look for a value in xml file
        :return fecha: datetime value of tag found
        :rtype fecha: Datetime
        """
        res = ""
        if not self.env.user.partner_id.tz:
            raise UserError(_("Set your time zone"))

        user_tz = timezone(self.env.user.partner_id.tz)
        try:
            fecha = parse(self._parse_xml(tag))
        except AttributeError as e:
            _logger.debug(e)
        else:
            fecha = user_tz.localize(fecha)
            fecha = fecha.astimezone(utc)
            res = fields.Datetime.to_string(fecha)

        return res

    @property
    def hash_type(self):
        """Easy way to select the kind of hash to use for get sello
        based on CFDI version
        """
        return NAMES[self.config_id.version]["hash_type"]

    @property
    def xslt_file(self):
        return NAMES[self.config_id.version]["xslt_file"]

    @api.multi
    def unlink(self):
        """Prevents user deletes a CFDI in signed or validate state
        """
        if any(self.filtered(lambda r: r.state in ("signed", "done"))):
            raise UserError(_("You can't delete signed CFDI!\nPlease cancel it first"))
        return super(IrAttachmentFacturaeMx, self).unlink()

    @api.multi
    def action_sign_sat(self):
        """Function use to sign XML in SAT

        This function using param_pacs parameter try to sign XML in SAT
        and creates XML ir.attachment related with ir.attachment.facturae.mx
        """
        # pylint: disable=broad-except
        self.ensure_one()
        record = self.env[self.type_attachment].browse(self.res_id)
        # Get all setup PACs in journal
        params_pacs = self.get_pacs()
        # Call function which connects with PACs but create a savepoint to avoid
        # loose given record state and prevent sending several times same record to
        # get PAC validation
        try:
            with self.env.cr.savepoint():
                res = params_pacs.sign_cfdi(self.file_xml.datas)
                xml_data = res.get("cfdi_xml", False).encode("UTF-8")
                # Create XML attachment
                attachment_id = self._create_ir_attachment(
                    self.name, "xml", xml_data, record.id
                )
                # Save created XML into ir.attachment.facturae.mx object
                self.write({"file_xml_sign": attachment_id, "state": "signed"})
                # Remove relation from not signed XML to current record
                self.file_xml.write({"res_id": 0, "res_model": "mail.compose.message"})
        except Exception as e:
            record.message_post(e.message)

    def _create_ir_attachment(self, name, extension, data, res_id):
        """Function use to create ir.attachment related

        @param name (string): Name of ir.attachment
        @param extentision (string): File extension
        @param data (string): Data to be loaded as string and
            without codification
        @param res_id (integer): Id
        @return: id's ir.attachment created
        """
        attachment_obj = self.env["ir.attachment"]
        attachment = attachment_obj.create(
            {
                "type": "binary",
                "name": ".".join([name, extension]),
                "datas_fname": ".".join([name, extension]),
                "datas": base64.encodestring(data),
                "res_model": self.type_attachment,
                "res_id": res_id,
            }
        )
        return attachment.id

    @api.multi
    def action_validate(self):
        """Validate current CFDI on SAT using this logic:

        * For CFDI's with no file_xml_sign create a new XML and sign it
        * For CFDI's with file_xml_sing ask SAT for document status
        """
        # Ensure this method is called only for one record
        self.ensure_one()

        if not self.state == "draft":
            raise AssertionError(
                _("`action_validate` must be called only for `draft` cfdis")
            )

        if self.config_id and not self.file_xml_sign:
            self.sign_cfdi_sat()
        else:
            self.validate_external_cfdi()

    @api.multi
    def action_create_pdf(self, raise_exception=True):
        """Create PDF for current CFDI record"""
        # pylint: disable=broad-except
        # Ensure this method is called only for one record
        self.ensure_one()

        record = self.env[self.type_attachment].browse(self.res_id)
        # Create PDF report
        try:
            data = self.create_report(
                record.id, self.config_id.template_pdf_sign, model=record._name
            )
        except Exception as e:
            _logger.error(_("Unexpected error creating PDF: {e}").format(e=e))
            if raise_exception:
                raise
        else:
            # Create PDF attachment
            attachment_id = self._create_ir_attachment(
                self.name, "pdf", data, record.id
            )
            # Save created attachment into ir.attachment.facturae.mx object
            self.write({"file_pdf": attachment_id, "state": "done"})

    @api.multi
    def action_send_mail(self, raise_exception=True):
        """Post new message into related record to send CFDI by mail"""
        # pylint: disable=broad-except
        # Ensure this method is called only for one record
        self.ensure_one()

        record = self.env[self.type_attachment].browse(self.res_id)

        # Create mail content
        _logger.debug("Start processing mail template")
        try:
            values = self.config_id.email_template_id.generate_email(record.id)[0]
            # Grab attachments
            values["attachment_ids"] = [self.file_xml_sign.id, self.file_pdf.id]
            values["type"] = "comment"
            values["subtype"] = "mail.mt_comment"
            # Add all object followers to mail vals to allow simplify
            # partner management and auto subscriptions
            partner_ids = values.get("partner_ids", [])
            partner_ids.extend([p.id for p in record.message_follower_ids])
            values["partner_ids"] = partner_ids
            # Send mail
            record.with_context(mail_post_autofollow=True).message_post(**values)
        except Exception as e:
            _logger.error(_("Unexpected error rendering email: {e}").format(e=e))
            if raise_exception:
                raise

    @api.multi
    def action_cancel(self):
        """Calcel current CFDI on SAT.
        * For CFDI's with incoming type use any available PAC for cancel it
        * For CFDI's with outgoing type do nothing
        """
        self.ensure_one()
        if self.cfdi_type == "incoming" and self.state in ("signed", "done", "waiting"):
            params_pacs = self.get_pacs()
            # Cancel function does not generate a new XML file
            # either sign any file using user certificates
            # the problem is that each PAC request CFDI information
            # in different way in order to cancel CFDI
            # Call function which connects with PACs
            cancel_answer = params_pacs.cancel_cfdi(self)

            if cancel_answer:
                self.cancel_cfdi()
            elif cancel_answer is None:
                self.write({"state": "waiting"})
            else:
                self.undo_waiting_state()
            return cancel_answer

    @api.multi
    def action_consult_cancellation_status(self):
        """Consult current cfdi cancel status on SAT"""
        self.ensure_one()

        consult_answer = None
        pacs = self.get_pacs()

        if self.state in ["waiting", "cancel"]:
            consult_answer = pacs.consult_cfdi_cancellation_status(self)
            if consult_answer:
                self.cancel_cfdi()
            elif consult_answer is None:
                pass
            else:
                self.undo_waiting_state()
        return consult_answer

    def sign_xml(self, xml_data):
        # Get current certificate for current company
        certificate = self.company_id.certificate_id
        if not certificate:
            raise UserError(
                _(
                    "There is no VALID certificate for company: {company}!\n"
                    "It is possible that the CERTIFICATE is already expired.\n"
                    "Go to Configuration >> Companies >> Companies to verify "
                    "it."
                ).format(company=self.company_id.name)
            )
        comprobante = "/cfdi:Comprobante"
        ns = {"cfdi": "http://www.sat.gob.mx/cfd/3"}

        # Write certificate number in XML
        node_comprobante = xml_data.xpath(comprobante, namespaces=ns)[0]
        node_comprobante.attrib["NoCertificado"] = certificate.serial_number

        cert_str = self._get_certificate_str(certificate.certificate_file_pem)
        if not cert_str:
            raise UserError(
                _(
                    "Can't get the Certificate Number of the voucher.\n"
                    "Ckeck your configuration."
                )
            )
        cert_str = cert_str.replace(" ", "").replace("\n", "")
        node_comprobante.attrib["Certificado"] = cert_str

        # Get path of XLST file
        xlst_file = pkg_resources.resource_filename(
            "openerp.addons.l10n_mx_ir_attachment_facturae", "cfd/" + self.xslt_file
        )
        # Transform XML into a string
        xml_string = ET.tostring(xml_data, encoding="utf-8")
        original_chain = self._xml2cad_orig(xml_string, xlst_file)

        sign_str = self._get_sello(
            certificate.certificate_key_file_pem, original_chain, self.hash_type
        )
        if not sign_str:
            raise UserError(
                _(
                    "Can't generate the stamp of the voucher.\n"
                    "Check your configuration!"
                )
            )

        # Write digital signature in XML
        node_comprobante.attrib["Sello"] = sign_str

        data_xml = ET.tostring(xml_data, encoding="utf-8")
        data_xml = data_xml.replace(
            '<?xml version="1.0" encoding="UTF-8"?>',
            '<?xml version="1.0" encoding="UTF-8"?>\n',
        )
        return data_xml

    def ensure_one_cfdi(self):
        """Verifies that there is only one cfdi for a given record. Raise an
        error otherwise.
        """
        cfdis = self.search_count(
            [
                ("res_id", "=", self.res_id),
                ("type_attachment", "=", self.type_attachment),
                ("company_id", "=", self.company_id.id),
                ("state", "in", ("signed", "done")),
            ]
        )
        if cfdis:
            raise ValidationError(
                _(
                    "Something really wrong happen!\n"
                    "There is already a signed CFDI for {object} "
                    "with id: {id}"
                ).format(object=self.type_attachment, id=self.res_id)
            )

    @staticmethod
    def _get_certificate_str(certifcate_pem):
        """ Function use to get certificate

        @param certificate_pem (string): string which contains certificate
            in PEM format
        @return: string
        """
        certificate = m2.X509.load_cert_string(certifcate_pem, 1)
        # Get certificate decode in base64 without \n
        certificate_base64 = base64.encodestring(certificate.as_der()).replace("\n", "")
        return certificate_base64

    @staticmethod
    def _get_sello(private_key, original_chain, hash_type):
        """ Function use to generate digital signature

        @param private_key (string): string containing private key
            in PEM format
        @param original_chain (string): Information use
            to create digital signature
        @param hash_type (string): hash type, it could be any value accepted
            for M2Crypto library as for example:
                sha1 or sha256
        @return: digital signature -> string
        """
        keys = m2.RSA.load_key_string(private_key)
        digest = hashlib.new(hash_type, original_chain).digest()
        digital_signature = base64.b64encode(keys.sign(digest, hash_type))
        return digital_signature

    @staticmethod
    def _xml2cad_orig(xml, xlst):
        """ Function use to get original chain

        @param data (string): string form of a XML from
            where original chain will be created
        @param xlst (string): path to the xlst file
        @return: original_chain -> string
        """
        xdoc = ET.fromstring(xml)
        xsl_root = ET.parse(xlst)
        xsl = ET.XSLT(xsl_root)
        original_chain = xsl(xdoc)
        return str(original_chain)

    def _parse_xml(self, attrib, element="*"):
        """Use self.parse_data to read xml related with given record parse it
        and return value for given attribute

        @param attrib (string): name of the attribute to looking for
        @param element (string): optional element to restrict attribute search
        """
        self.ensure_one()
        return self.parse_data(
            self.file_xml_sign.index_content, attrib, element=element
        )

    @classmethod
    def parse_data(cls, data, attrib, element="*", namespaces=None):
        """Return value for given attribute

        This function looks for an attribute named as attrib parameter into
        a given string that represents an XML file and then returns its
        value or returns False

        @param data (string): string to parse
        @param attrib (string): name of the attribute to looking for
        @param element (string): optional element to restrict attribute search
        @param namespaces (dictionary): optional namespaces to pass to the
            find function. If not provided then NAMESPACES will be used
        @return: string
        """
        namespaces = namespaces or NAMESPACES
        if not data:
            return ""
        root = parse_xml_string(data)
        complete_pattern = "".join([".//", element, "[@", attrib, "]"])
        result = root.find(complete_pattern, namespaces=namespaces)
        # if not element is because attribute is in the first node
        # and regular expression does not match
        if result is None:
            complete_pattern = "".join([".[@", attrib, "]"])
            result = root.find(complete_pattern, namespaces=namespaces)
        value = result.attrib[attrib]
        return value

    @api.multi
    def create_report(self, attach_id, report_name, model):
        """ Create report

        @param attach_id: Id of the record from which values will be taken
        @type attach_id: integer
        @param model: Model name
        @type model: string
        @param report_name: Name of report template
        @type report_name: string
        """
        result, dummy = render_report(
            self.env.cr, self.env.uid, [attach_id], report_name, {"model": model}
        )
        return result

    @api.multi
    def create_report_adenda(self):
        """ Create Addenda for XML """
        self.ensure_one()
        record = self.env[self.type_attachment].browse(self.res_id)
        adenda = self.cfdi_adenda_id
        result = ""
        if adenda:
            report_name = adenda.report_id.report_name
            result = self.create_report(record.id, report_name, record._name)
        _logger.debug("Addenda before attach to XML: %s", result.decode("UTF-8"))
        return result

    @api.multi
    def create_report_xml(self):
        """ Create XML report and sign with company certificates"""
        self.ensure_one()
        config = self.config_id
        record = self.env[self.type_attachment].browse(self.res_id)
        # Render XML into string
        result = self.create_report(record.id, config.template_xml_sign, record._name)
        # Adjunta Addenda al XML
        if self.cfdi_adenda_id:
            adenda = self.cfdi_adenda_id
            result_adenda = self.create_report_adenda()
            document = minidom.parseString(result)
            node_comprobante = document.documentElement
            a_document = minidom.parseString(result_adenda.strip())
            adenda_content = a_document.documentElement

            if adenda.xml_prefix and adenda.xml_ns and adenda.xml_xsd:
                # Si estan definidos los campos de Namespace,
                #  entonces es un cfdi:Complemento
                node_name = "cfdi:Complemento"
                xsi_schemans = "http://www.w3.org/2001/XMLSchema-instance"
                ns_xmlns = "http://www.w3.org/2000/xmlns/"
                ns_xsd_sat = node_comprobante.getAttribute("xsi:schemaLocation")
                # Se agregan los Namespaces
                node_comprobante.setAttributeNS(
                    xsi_schemans,
                    "xsi:schemaLocation",
                    "%s %s %s" % (ns_xsd_sat, adenda.xml_ns, adenda.xml_xsd)
                )
                node_comprobante.setAttributeNS(
                    ns_xmlns, adenda.xml_prefix, adenda.xml_ns
                )
            else:
                # Caso contrario, entonces es una cfdi:Addenda
                node_name = "cfdi:Addenda"

            node_adenda = document.createElement(node_name)
            node_adenda.appendChild(adenda_content)
            node_comprobante.appendChild(node_adenda)
            result = document.toxml(encoding="UTF-8")

        xml_data = parse_xml_string(result.strip())
        xml_data = self.sign_xml(xml_data=xml_data)
        # Create XML attachment
        attachment_id = self._create_ir_attachment(
            self.name, "xml", xml_data, record.id
        )
        # In case we are recreating the file, delete relation from old XML and record
        if self.file_xml:
            self.file_xml.res_id = False
        self.write({"file_xml": attachment_id})

    @api.one
    def sign_cfdi_sat(self):
        """ Function use to create and sign a CFDI

        The idea is from whatever object you are, for example invoice,
        if you want to have your invoice sign in SAT only call this
        function.
        This function performs next steps in same order
            1.- Create a XML using template set in ir.attachment.mx.config
            2.- Sign XML using company certificates
            2.- Get UUID for XML from SAT by sending to active PACs
            3.- Create a PDF using template set in ir.attachment.mx.config
            4.- Post message on main record based on template set on
                ir.attachment.mx.config

        Result is saved only in this ir.attachmet.facturae.mx object
        Other objects have to connect to this model to get desired information
        about result, for example to get uuid from invoice
        invoice have to connect with the related ir.attachmet.facturae.mx and
        read uuid field

        @param object_id (integer): database id of the object
        @param object_model (string): model name
        """
        # Validates there is no other CFDI signed for this invoice
        self.ensure_one_cfdi()
        config = self.config_id

        # Create XML and sign with company certificates
        if not self.file_xml or self.env.context.get("recreate_xml", False):
            self.create_report_xml()

        # Send XML to PAC for get SAT signature
        if self.state == "draft":
            self.action_sign_sat()

        # Create PDF report
        if self.state == "signed":
            self.action_create_pdf(raise_exception=False)

        # Post a message into the main record that will send CFDI by mail
        if config.email_template_id and self.state == "done":
            self.action_send_mail(raise_exception=False)
        return

    @api.model
    def get_pacs(self):
        """get all setup PACs in journal"""
        record = self.env[self.type_attachment].browse(self.res_id)
        # Get all setup PACs in journal
        return record.company_id.pac_ids

    @api.multi
    def undo_waiting_state(self):
        """when cancel is negate, undo waiting state"""
        self.ensure_one()
        if self.state in "waiting":
            if self.file_pdf:
                self.write({"state": "done"})
            else:
                self.write({"state": "signed"})

    @api.model
    def cancel_cfdi(self):
        """ Cancel CFDI and rewrite PDF

             Since we do not generate a new XML
             Only generate PDF file
             the idea is to overwrite the existing one
        """
        record = self.env[self.type_attachment].browse(self.res_id)

        data = self.create_report(
            record.id, self.config_id.template_pdf_sign, model=record._name
        )
        # Like said before, instead of creating new attachment
        # overwrite existing one
        self.file_pdf.write({"datas": base64.encodestring(data)})
        self.write({"cfdi_fecha_cancelado": fields.Datetime.now(), "state": "cancel"})

    @api.multi
    def validate_external_cfdi(self):
        """Validate an XML from other system
        """
        self.ensure_one()
        result = self.validate_cfdi_folio_fiscal()
        # If webservice does not return 'Vigente' state raises an exception
        if result != "Vigente":
            raise ValidationError(
                _(
                    "The CFDI provided did not passed the SAT validation\n"
                    "The UUID sate on SAT servers is %s"
                )
                % result
            )
        # Update status for this CDFI according to response
        self.write({"state": "done"})

    @api.multi
    def validate_cfdi_folio_fiscal(self):
        """Consume SAT service for get status for a given CFDI"""
        self.ensure_one()
        # Webservice parameters
        url = (
            "https://consultaqr.facturaelectronica.sat.gob.mx/ConsultaCFDIService.svc"
        )
        headers = {
            "SOAPAction": "http://tempuri.org/IConsultaCFDIService/Consulta",
            "Content-Type": "text/xml; charset=utf-8",
        }
        namespaces = {
            "a": "http://schemas.datacontract.org/2004/07/Sat.Cfdi.Negocio.ConsultaCfdi"
                 ".Servicio",
        }
        data = SOAPENV_SAT_VALIDATE.format(
            # Not sure why we need escape twice the TID
            re=escape(escape(self.rfc_emisor)),
            rr=escape(escape(self.rfc_receptor)),
            tt=self.total,
            id=self.uuid,
        )
        req = Request("POST", url, data=data, headers=headers)
        # Consume webservice for validate uuid
        response = Session().send(req.prepare(), timeout=60)
        try:
            response.raise_for_status()
        except HTTPError:
            # Error communicating with SAT service
            raise ValidationError(response.text.encode("utf-8"))

        # Parse webservice response
        response = ET.fromstring(response.content)
        result = response.find(".//a:Estado", namespaces=namespaces).text
        return result
