# -*- coding: utf-8 -*-
# Copyright 2019 OpenPyme
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl).

import base64
import hashlib

import M2Crypto as m2  # noqa

from openerp import api, fields, models

from ..lib.utils import PkiUtils


class ResCompanyFacturaeFiel(models.Model):

    _name = "res.company.facturae.fiel"
    _description = "Company Facturae Fiel"

    _rec_name = "serial_number"

    @api.model
    def _default_company_id(self):
        res_company_obj = self.env["res.company"]
        company = res_company_obj._company_default_get("res.company.facturae.fiel")
        return company

    company_id = fields.Many2one(
        "res.company",
        "Company",
        required=True,
        default=lambda self: self._default_company_id(),
        help="Company where you add this fiel",
    )
    fiel_file = fields.Binary(
        string="file of fiel",
        filters="*.cer,*.certificate,*.cert",
        required=True,
        help="File .cert of the FIEL",
    )
    fiel_key_file = fields.Binary(
        string="file KEY of fiel",
        filters="*.key",
        required=True,
        help="File .key of the FIEL",
    )
    fiel_password = fields.Char(
        string="Password of fiel",
        size=64,
        invisible=False,
        required=True,
        help="Password for login to SAT",
    )
    fiel_file_pem = fields.Binary(
        string="file of fiel PEM",
        filters="*.pem,*.cer,*.certificate,*.cert",
        help="This file is generated with the file.cer",
    )
    fiel_key_pem = fields.Binary(
        string="file KEY of fiel PEM",
        filters="*.pem,*.key",
        help="This file is generated with the file.key",
    )
    date_start = fields.Date(help="Date start the fiel before the SAT")
    date_end = fields.Date(help="Date end of validity of the fiel")
    serial_number = fields.Char(help="Number of serie of the fiel")
    active = fields.Boolean(default=True, help="Indicate if this fiel is active")

    @api.model
    @api.returns("self", lambda value: value.id)
    def create(self, vals):
        res = super(ResCompanyFacturaeFiel, self).create(vals)
        company = self.env["res.company"]
        company._get_info(res)
        return res

    def _load_certificate_and_private_key(self, vals):
        """Verify fiel and key files to be loaded are right
        If they are right return files in der format to be parsed
        """
        pkutils = PkiUtils(vals.fiel_file, vals.fiel_key_file, vals.fiel_password)
        pkutils.validate_cert_key()
        cer_pem = pkutils.get_cert_pem()
        key_pem = pkutils.get_key_pem()
        # Save fiel as PEM and KEY as PEM too
        vals.write({"fiel_file_pem": cer_pem, "fiel_key_pem": key_pem})
        return pkutils.get_cer_der()

    @api.multi
    def _get_fiel(self):
        """
        this function gets the FIEL associated with the company.
        :return: FIEL of the comapany.
        """
        self.ensure_one()
        company_id = self.env.user.company_id
        fiel = company_id.fiel_id
        return fiel

    @api.multi
    def _firmar_xml(self, xml):
        """
        function to sign XML in string format.
        :param xml: XML in string format.

        :return: XML information signed and encoded to base64.
        """
        self.ensure_one()
        keys = m2.RSA.load_key_string(self.fiel_key_pem)
        digest = hashlib.new("sha1", xml).digest()
        digital_signature = base64.b64encode(keys.sign(digest, "sha1"))
        return digital_signature

    @api.multi
    def _certificate_base64(self):
        """
        function that encodes base64 the faithful certificate in PEM format.

        :return: certificate in base64 format.
        """
        cert = m2.X509.load_cert_string(self.fiel_file_pem, 1)
        certificate = base64.b64encode(cert.as_der())
        return certificate

    @api.multi
    def _cretificate_issuer(self):
        cert = m2.X509.load_cert_string(self.fiel_file_pem, 1)
        certificate = cert.get_issuer()
        certificate = certificate.as_text()
        return certificate
