# -*- coding: utf-8 -*-

import logging

from openerp import api, fields, models
from openerp.exceptions import ValidationError, Warning as UserError
from openerp.tools.translate import _

from . import exceptions


_logger = logging.getLogger(__name__)


class ParamsPac(models.Model):
    _name = "params.pac"
    _order = "sequence ASC"

    @classmethod
    def get_driver_fc_sign(cls):
        """function to inherit from custom PAC module"""
        return {}

    @classmethod
    def get_driver_fc_cancel(cls):
        """function to inherit from custom PAC module"""
        return {}

    @classmethod
    def get_driver_fc_status(cls):
        """function to inherit from custom PAC module"""
        return {}

    name = fields.Char(size=128, required=True, help="Name for this PAC")
    user = fields.Char(size=128, help="Name user for login to PAC")
    password = fields.Char(size=128, help="Password user for login to PAC")
    pac_type = fields.Selection([], type="char", required=True)
    company_id = fields.Many2one(
        "res.company",
        "Company",
        required=True,
        default=lambda self: self.env.user.company_id,
        help="Company where this PAC will be used",
    )
    active = fields.Boolean(default=1, help="Indicate if this PAC is active")
    sequence = fields.Integer(
        default=10,
        help="If have more than one PAC is active, the one with lower number"
        "in sequence will be used first.",
    )
    certificate_link = fields.Char(
        size=256,
        help="PAC public certificate used to validate authenticity for a given"
        "XML file.",
    )
    url_webservice_sign = fields.Char(
        "URL WebService",
        size=256,
        required=True,
        help="URL of WebService used for send to sign the XML to PAC",
    )
    namespace_sign = fields.Char(
        "NameSpace",
        size=256,
        help="NameSpace of XML of the page of WebService of the PAC",
    )
    url_webservice_cancel = fields.Char(
        "URL WebService",
        size=256,
        required=True,
        help="URL of WebService used for send to sign the XML to PAC",
    )
    namespace_cancel = fields.Char(
        "NameSpace",
        size=256,
        help="NameSpace of XML of the page of WebService of the PAC",
    )

    @api.multi
    def fire_connection(self, data, action):
        """
        This function handles connection to the PACs
        All PACs subscribe using callback to the
        get_driver_fc_sign or  get_driver_fc_cancel function

        When some PAC is not available then next PAC is tried
        That's why the reason of pac_ids parameter
        @param self : All possible PACs to be used
        @param data (string):
            base64 encoding string in case of get_driver_fc_sign
            uuid in case of get_driver_fc_cancel
        @param action (string): action to perform it could
            either get_driver_fc_sign or get_driver_fc_cancel
        @return: Dictionary
        """
        # The order is important because it means
        # the priority selected by the user
        # If all PACs would fail this variable will recollect
        # all error in order to display them at the end
        pacs_errors = []
        message_error = ""
        for param_pac in self:
            # Get all possible drivers to perform action
            # we done this here to ensure function is associated only with one pac
            functions = getattr(param_pac, action)()
            try:
                result = functions[param_pac.pac_type](data)
            # This kind of exception means that
            # nevertheless you try with all PACs in the world
            # CFDI would be failed
            except (
                    exceptions.ValidationError, exceptions.PrimarySectorError,
                    exceptions.CancelError
            ) as error:
                _logger.debug(_(repr(error)))
                raise UserError(
                    _(repr(error)),
                )
            # TODO: We need to catch errors like timeout and similar
            # because the idea is when a PAC fails try next
            # pylint: disable=broad-except
            except Exception as error:
                # If something went wrong try with next PAC
                message_error = "\n".join(
                    [param_pac.name, repr(error)],
                )
                pacs_errors.append(message_error.encode("UTF-8") + "\n")
                _logger.info(
                    message_error.encode("UTF-8"),
                )
                continue
            # api.one decorator always returns a list
            # and some get_driver_fc_sign functions are decorated with api.one
            if isinstance(result, list):
                result = result[0]
            # Some PAC get right result
            break
        try:
            return result
        except UnboundLocalError:
            # If result value does not exist
            # it is because all PACs failed
            raise UserError(repr(pacs_errors))

    @api.multi
    def sign_cfdi(self, data):
        """
        Sign_cfdi and cancel_cfdi perform both similar actions
        Both connect to PAC, if PAC fails code messages are the same
        and so on.

        @param self : All possible PACs to be used
        @param data (string): data to sign in base64 encoding
        @return: Dictionary
        """
        self.check_pacs()
        result = self.fire_connection(data, "get_driver_fc_sign")
        return result

    @api.multi
    def cancel_cfdi(self, cfdi):
        """
        Sign_cfdi and cancel_cfdi perform both similar actions
        Both connect to PAC, if PAC fails code messages are the same
        and so on.

        :param self : All possible PACs to be used
        :param cfdi: cfdi to cancel
        :type cfdi: `ir.attachment.facturae.mx`
        :return: Return true if the cfdi can be canceled, false if it can't be canceled
            and None if must wait to cancel
        """
        self.check_pacs()
        result = self.fire_connection(cfdi, "get_driver_fc_cancel")
        return result

    @api.multi
    def consult_cfdi_cancellation_status(self, cfdi):
        """Connect to PAC to check cfdi cancel status
        :param cfdi: cfdi to review
        :type cfdi: ir.attachment.facturae.mx
        """
        self.check_pacs()
        _logger.info("Consulting status for uuid: %s", cfdi.uuid)
        result = self.fire_connection(cfdi, "get_driver_fc_status")
        return result

    @api.multi
    def check_pacs(self):
        # pylint: disable=len-as-condition
        """Verify if there are pacs selected"""
        if not len(self):
            raise ValidationError(
                _("No PAC have being selected.\nGo to Configuration "
                  ">> Accounting >> Electronic invoicing (MX) and select "
                  "a PAC."),
            )
