# -*- coding: utf-8 -*-
# pylint: disable=consider-merging-classes-inherited

import base64
import logging
import re
import string
from requests import Request, Session

import xmltodict

from openerp import api, fields, models
from openerp.exceptions import Warning as UserError
from openerp.tools.translate import _

from ..lib import xpd_envelopes
from ..lib.pac_answer import PacConsultAnswer
from ..models import exceptions as pac_exceptions


_logger = logging.getLogger(__name__)

TIMEOUT = 60

_ACTIONS = {"timbrar": "timbrar", "cancelar": "", "consultar": ""}


class ParamsPac(models.Model):
    _inherit = "params.pac"

    pac_type = fields.Selection(
        selection_add=[("pac_facturalo", _("Expide tu factura"))]
    )

    def get_driver_fc_sign(self):
        """
        Implement driver for sign invoices
        """
        factura_mx_type__fc = super(ParamsPac, self).get_driver_fc_sign()
        if factura_mx_type__fc is None:
            factura_mx_type__fc = {}
        factura_mx_type__fc.update({"pac_facturalo": self.sign_file})
        return factura_mx_type__fc

    def get_driver_fc_cancel(self):
        """
        Implement driver for cancel invoices
        """
        factura_mx_type__fc = super(ParamsPac, self).get_driver_fc_cancel()
        if factura_mx_type__fc is None:
            factura_mx_type__fc = {}
        factura_mx_type__fc.update({"pac_facturalo": self.cancel_xpd_cfdi})
        return factura_mx_type__fc

    def get_driver_fc_status(self):
        """ Implement driver for consult cfdi's"""
        factura_mx_type__fc = super(ParamsPac, self).get_driver_fc_status()
        if factura_mx_type__fc is None:
            factura_mx_type__fc = {}
        factura_mx_type__fc.update({"pac_facturalo": self.consult_xpd_cfdi})
        return factura_mx_type__fc

    @classmethod
    def validate_response(cls, soap_response):
        """if response.status_code not 200 then raise Exception"""
        if str(soap_response.status_code) not in "200":
            raise Exception(
                "Bad response: " + str(soap_response.status_code)
                + "," + soap_response.content
            )

    @api.one
    def sign_file(self, fdata=None):
        """
        @params
            fdata : File.xml base64 codificated
        """
        msg = ""
        if self.user == "pruebas":
            msg += _("WARNING, SIGNED IN TEST!!!!\n\n")

        data = xpd_envelopes._SOAPENV % (
            self.user,
            self.password,
            fdata,
        )
        _logger.debug(data)
        headers = {
            "SOAPAction": "'timbrar'",
            "Content-Type": "text/xml;charset=UTF-8",
        }
        s = Session()
        req = Request("POST", self.url_webservice_sign, data=data, headers=headers)
        prepped = req.prepare()
        try:
            # Timbrar factura
            response = s.send(prepped, timeout=TIMEOUT)
        except Exception as e:
            # Error al establecer comunicación con el PAC
            _logger.debug(e)
            raise e
        else:
            self.validate_response(response)
        res = xmltodict.parse(response.text)
        _logger.debug(res)
        # Procesar los resultados obtenidos del PAC
        try:
            resultado = res["soapenv:Envelope"]["soapenv:Body"]["ser:timbrarResponse"][
                "return"
            ]
        except KeyError as e:
            resultado = res["soap:Envelope"]["soap:Body"]["ns2:timbrarResponse"][
                "return"
            ]
            _logger.debug(e)
        except Exception as e:
            _logger.debug(e)
            raise UserError(_("Unknow answer.\n%s") % resultado)

        mensaje = resultado["mensaje"]
        valid_codes = ["200", "504"]
        if resultado["codigo"] in valid_codes:
            # Proceso satisfactorio
            timbre = resultado["timbre"]
            uuid = resultado.get("uuid", "")
            if not uuid:
                regex = "<cfdi:Complemento>(.*?)</cfdi:Complemento>"
                # Obtiene el folio de la factura
                try:
                    timbref = re.search(regex, resultado["timbre"]).group(1)
                    uuid = re.search('UUID="(.+?)" ', timbref).group(1)
                except:  # noqa
                    uuid = ""

            # Get back XML information to file before store it
            timbre = '<?xml version="1.0" encoding="UTF-8"?>\n' + timbre
            # Codifica el XML para almacenarlo
            f = base64.encodestring(timbre.encode("UTF-8") or "")

            msg += string.join([mensaje, ". Folio Fiscal ", uuid, "."])
            return {"file": f, "msg": msg, "cfdi_xml": timbre}
        elif resultado["codigo"] in pac_exceptions.validation_error_list:
            # El CFDI no ha sido timbrado
            raise pac_exceptions.ValidationError(
                code=resultado["codigo"], message=mensaje
            )
        else:
            raise Exception(" ".join(["Code", resultado["codigo"], mensaje]))

    @api.model
    def consult_xpd_cfdi(self, cfdi):
        """ Consul the cfdi status with XPD pac

        :param uuid: UUID of CFDI to consult
        """
        method = "consultar"
        try:
            data = xpd_envelopes._SOAPENV_CONSULTAR % (
                self.user,
                self.password,
                cfdi.no_certificado.encode("utf-8"),
                cfdi.rfc_emisor.encode("utf-8"),
                cfdi.rfc_receptor.encode("utf-8"),
                cfdi.total,
                cfdi.uuid.encode("utf-8").lower(),
            )
        except Exception as e:
            e.message = "Error al crear ENVELOPE"
            _logger.info(e)
            raise e
        else:
            _logger.debug(data)
        headers = self.create_header(method)
        s = Session()
        req = Request("POST", self.url_webservice_cancel, data=data, headers=headers)
        prepped = req.prepare()
        try:
            # Consulta el estado de la factura
            response = s.send(prepped, timeout=TIMEOUT)
        except Exception as e:
            # Error al establecer comunicación con el PAC
            _logger.debug(e)
            raise e
        else:
            self.validate_response(response)
        res = xmltodict.parse(response.text)
        _logger.debug(res)

        # Procesar los resultados obtenidos del PAC
        try:
            resultado = res["soapenv:Envelope"]["soapenv:Body"][
                "ns2:consultarCfdiResponse"
            ][
                "ns2:consultarCfdiResponse"
            ]
        except KeyError as e:
            resultado = res["soapenv:Envelope"]["soapenv:Body"][
                "ns2:consultarCfdiResponse"
            ][
                "ns2:consultarCfdiResponse"
            ]
            _logger.debug(e)
        except Exception as e:
            _logger.debug(e)
            raise UserError(_("Unknow answer.\n%s") % resultado)

        res = resultado["ns2:responseGenericoQr"]["ns2:servicio"]
        res = res.popitem(True)
        res = res[1]
        res = xmltodict.parse(res.encode("utf-8"))
        resultado = res["s:Envelope"]["s:Body"]["ConsultaResponse"]["ConsultaResult"]
        estatus_cancelacion = resultado["a:EstatusCancelacion"]
        estado_cfdi = resultado["a:Estado"]
        consult_answer = PacConsultAnswer(cfdi.uuid, estado_cfdi, estatus_cancelacion)
        return consult_answer.check_cancel_status()

    @api.model
    def cancel_xpd_cfdi(self, cfdi):
        """ Cancel CFDI

        :param uuid: uuid of cfdi to cancel
        :type uuid: string
        """
        method = "cancelar"
        try:
            data = xpd_envelopes._SOAPENV_CANCEL % (
                self.user,
                self.password,
                cfdi.no_certificado.encode("utf-8"),
                cfdi.rfc_emisor.encode("utf-8"),
                cfdi.rfc_receptor.encode("utf-8"),
                cfdi.total,
                cfdi.uuid.encode("utf-8").lower(),
            )
        except Exception as e:
            e.message = "Error while create ENVELOPE"
            _logger.info(e)
            raise e
        else:
            _logger.debug(data)
        headers = self.create_header(method)
        s = Session()
        req = Request("POST", self.url_webservice_cancel, data=data, headers=headers)
        prepped = req.prepare()
        try:
            # Cancela factura
            response = s.send(prepped, timeout=TIMEOUT)
        except Exception as e:
            # Error al establecer comunicación con el PAC
            _logger.debug(e)
            raise e
        else:
            self.validate_response(response)
            _logger.debug(response.text)
        res = xmltodict.parse(response.text)
        _logger.debug(res)
        # Procesar los resultados obtenidos del PAC
        try:
            resultado = res["soapenv:Envelope"]["soapenv:Body"][
                "ns2:cancelarCfdiResponse"
            ][
                "ns2:cancelarCfdiResponse"
            ]
        except KeyError as e:
            resultado = res["soapenv:Envelope"]["soapenv:Body"][
                "ns2:cancelarCfdiResponse"
            ][
                "ns2:canselarCfdiResponse"
            ]
            _logger.debug(e)
        except Exception as e:
            _logger.debug(e)
            raise UserError(_("Unknow answer.\n%s") % resultado)
        mensaje = resultado["ns2:mensaje"]
        codigo = resultado["ns2:codigo"]
        # Procesar los resultados obtenidos del PAC
        if codigo not in ("200", "201", "202"):
            raise Exception(" ".join(["Code", codigo, mensaje]))
        return self.consult_xpd_cfdi(cfdi)
