# -*- coding: utf-8 -*-

import logging
from suds.client import (
    BuildError,
    Client,
    MethodNotFound,
    PortNotFound,
    ServiceNotFound,
    SoapHeadersNotPermitted,
    TypeNotFound,
    WebFault,
)

from openerp import fields, models
from openerp.exceptions import Warning as UserError
from openerp.tools.translate import _

from ..lib.pac_answer import PacConsultAnswer
from ..models import exceptions as pac_exceptions


_logger = logging.getLogger(__name__)


# Finkok does not return message errors
# that's why I created a dictionary with all errors
# described in its manual
_cancel_errors = {
    "300": "Usuario y contraseña inválidos",
    "203": "No corresponde el RFC del Emisor y de quien solicita la cancelación",
    "205": "UUID No existe",
    "704": "Error con la contraseña de la llave Privada",
    "708": "No se pudo conectar al SAT",
    "'711": "Error con el certificado al cancelar",
    "712": "El número de noCertificado es diferente al del número de"
    "certificado del atributo certificado",
}

TIMEOUT = 60
_ACTIONS = {"timbrar": "", "cancelar": "", "consultar": '"get_sat_status"'}


class ParamsPac(models.Model):
    _inherit = "params.pac"

    pac_type = fields.Selection(
        selection_add=[("pac_finkok", _("Finkok"))]
    )

    def get_driver_fc_sign(self):
        """
        Implement driver for sign CFDIs
        """
        factura_mx_type__fc = super(ParamsPac, self).get_driver_fc_sign()
        if factura_mx_type__fc is None:
            factura_mx_type__fc = {}
        factura_mx_type__fc.update({"pac_finkok": self.sign_cfdi_finkok})
        return factura_mx_type__fc

    def get_driver_fc_cancel(self):
        """
        Implement driver for cancel CFDIs
        """
        factura_mx_type__fc = super(ParamsPac, self).get_driver_fc_cancel()
        if factura_mx_type__fc is None:
            factura_mx_type__fc = {}
        factura_mx_type__fc.update({"pac_finkok": self.cancel_cfdi_finkok})
        return factura_mx_type__fc

    def get_driver_fc_status(self):
        """
        Implement driver for consult CFDIs
        """
        factura_mx_type__fc = super(ParamsPac, self).get_driver_fc_status()
        if factura_mx_type__fc is None:
            factura_mx_type__fc = {}
        factura_mx_type__fc.update({"pac_finkok": self.consult_cfdi_finkok})
        return factura_mx_type__fc

    @classmethod
    def create_header(cls, method):
        return {
            "SOAPAction": "%s" % (_ACTIONS[method]),
            "Content-Type": "text/xml; charset=utf-8"
        }

    def sign_cfdi_finkok(self, fdata):
        """Function uses to sign CFDIs in Finkok webservices

        @param fdata (string): String which contains XML in base64 format
        @return: dictionary: dictionary with only one key (cfdi_xml).
        The value of the key is a unicode which represents signed XML,
        this XML already has TimbreFiscalDigital Node with UUID value
        """
        # First read values from PAC to open a connection with Finkok
        # web services
        user, password = self.user, self.password
        server_url = self.url_webservice_sign
        try:
            server = Client(server_url, cache=None)
            # I thought it was necessary to create at first
            # a Soap request but only passing XML in base64 it works
            contenido = server.service.stamp(fdata, user, password)
        # Exceptions could be MethodNotFound, PortNotFound
        # ServiceNotFound, TypeNotFound, BuildError, SoapHeadersNotPermitted
        # and WebFault, we only catch the generic one.
        except (
            MethodNotFound,
            PortNotFound,
            ServiceNotFound,
            TypeNotFound,
            BuildError,
            SoapHeadersNotPermitted,
            WebFault,
        ) as e:
            raise e
        # If there is Incidencias attribute it is because
        # process failed
        if contenido.Incidencias:
            incidencia = contenido.Incidencias.Incidencia[0]
            # We already know that we only could have one item into the list
            menssage_error = incidencia.MensajeIncidencia
            code_error = incidencia.CodigoError
            if code_error in pac_exceptions.validation_error_list:
                raise pac_exceptions.ValidationError(
                    code=code_error, message=menssage_error
                )
            else:
                raise Exception(" ".join([code_error, menssage_error]))
        # I think, if there is not any Incidencia it is because
        # all worked well, then I have my XML properly signed in SAT
        else:
            return {"cfdi_xml": contenido.xml}

    def cancel_cfdi_finkok(self, cfdi):
        """Funcion uses to cancel a CFDI in SAT server
        using Finkok webservices

        @param UUID (string): Folio Fiscal of CFDI to be cancel
        """
        # First read values from PAC to open a connection with Finkok
        # web services
        user, password = self.user, self.password
        server_url = self.url_webservice_cancel
        # Get default company in order to get certificate
        certificate = self.company_id.certificate_id
        if not certificate:
            raise UserError(
                _(
                    "There is no VALID certificate for company: {company}!\n"
                    "It is possible that the CERTIFICATE is already expired.\n"
                    "Go to Configuration >> Companies >> Companies to verify "
                    "it."
                ).format(company=self.company_id.name)
            )
        company_rfc = self.company_id.partner_id.vat_split
        cer_file = certificate.certificate_file_pem.encode("base64")
        key_file = certificate.certificate_key_file_pem.encode("base64")
        cfdis = [cfdi.uuid]

        try:
            server = Client(server_url)
            # I thought it was necessary to create at first
            # a Soap request but only passing XML in base64 it works
            # The next lines are needed by the python suds library
            cfdis_list = server.factory.create("UUIDS")
            cfdis_list.uuids.string = cfdis
            result = server.service.cancel(
                cfdis_list, user, password, company_rfc, cer_file, key_file
            )
        # Exceptions could be MethodNotFound, PortNotFound
        # ServiceNotFound, TypeNotFound, BuildError, SoapHeadersNotPermitted
        # and WebFault, we only catch the generic one.
        except (
            MethodNotFound,
            PortNotFound,
            ServiceNotFound,
            TypeNotFound,
            BuildError,
            SoapHeadersNotPermitted,
            WebFault,
        ) as e:
            raise e
        # Get results, result is always in a list
        # but this list has only one value since we only send one
        # folio fiscal
        folio = result.Folios.Folio[0]
        if folio.EstatusUUID not in ["200", "201", "202", "no_cancelable"]:
            raise pac_exceptions.CancelError(
                folio.EstatusUUID, _cancel_errors.get(folio.EstatusUUID, "")
            )
        return self.consult_cfdi_finkok(cfdi)

    def consult_cfdi_finkok(self, cfdi):
        username = self.user
        password = self.password
        emisor = cfdi.rfc_emisor
        receptor = cfdi.rfc_receptor
        total = cfdi.total
        client = Client(self.url_webservice_cancel, cache=None)
        res = client.service.get_sat_status(
            username, password, emisor, receptor, cfdi.uuid, total
        )
        folio = res.sat
        estatus_cancelacion = folio["EstatusCancelacion"] or ""
        estado_cfdi = folio["Estado"]
        consult_answer = PacConsultAnswer(cfdi.uuid, estado_cfdi, estatus_cancelacion)
        return consult_answer.check_cancel_status()
