# -*- coding: utf-8 -*-
# pylint: disable=consider-merging-classes-inherited

import base64
import codecs
import logging
import time
import xml.dom.minidom
from requests import Request, Session

import xmltodict
from SOAPpy import WSDL
from SOAPpy.Errors import HTTPError, RecursionError, UnderflowError, UnknownTypeError

from openerp import fields, models, tools
from openerp.exceptions import Warning as UserError
from openerp.tools.translate import _

from ..lib import sf_envelopes
from ..lib.pac_answer import PacCancelAnswer, PacConsultAnswer
from ..models import exceptions as pac_exceptions


logger = logging.getLogger(__name__)

TIMEOUT = 60

_ACTIONS = {
    "timbrar": "",
    "cancelar": "",
    "consultar": '"urn:getStatusCancelacionAsincrona"'
}


class ParamsPac(models.Model):
    _inherit = "params.pac"

    pac_type = fields.Selection(
        selection_add=[("pac_sf", _("Solución Factible"))]
    )

    def get_driver_fc_sign(self):
        factura_mx_type__fc = super(ParamsPac, self).get_driver_fc_sign()
        if factura_mx_type__fc is None:
            factura_mx_type__fc = {}
        factura_mx_type__fc.update({"pac_sf": self._upload_ws_file})
        return factura_mx_type__fc

    def get_driver_fc_cancel(self):
        factura_mx_type__fc = super(ParamsPac, self).get_driver_fc_cancel()
        if factura_mx_type__fc is None:
            factura_mx_type__fc = {}
        factura_mx_type__fc.update({"pac_sf": self.sf_cancel})
        return factura_mx_type__fc

    def get_driver_fc_status(self):
        factura_mx_type__fc = super(ParamsPac, self).get_driver_fc_status()
        if factura_mx_type__fc is None:
            factura_mx_type__fc = {}
        factura_mx_type__fc.update({"pac_sf": self.sf_consult})
        return factura_mx_type__fc

    @classmethod
    def create_header(cls, method):
        return {
            "SOAPAction": "%s" % (_ACTIONS[method]),
            "Content-Type": "text/xml;charset=UTF-8",
        }

    def sf_cancel(self, cfdi):
        vat_emmiter = self.company_id.partner_id.vat_split
        # Get values necessary values from PAC configuration
        user = self.user
        password = self.password
        wsdl_url = self.url_webservice_cancel
        namespace = self.namespace_cancel
        wsdl_client = WSDL.SOAPProxy(wsdl_url, namespace)

        # Get certificate and key from certificate config in
        # company
        cer_csd = base64.encodestring(
            self.company_id.certificate_id.certificate_file_pem
        )
        key_csd = base64.encodestring(
            self.company_id.certificate_id.certificate_key_file_pem
        )
        contrasena_csd = self.company_id.certificate_id.certificate_password
        # It does not make sense to send an email
        # when the CFDI is cancelled since we have an ERP
        # and could review the cancelled CFDI directly in the ERP
        email_company = ""
        params = [
            user,
            password,
            cfdi.uuid,
            vat_emmiter,
            email_company,
            cer_csd,
            key_csd,
            contrasena_csd,
        ]
        wsdl_client.soapproxy.config.dumpSOAPOut = 0
        wsdl_client.soapproxy.config.dumpSOAPIn = 0
        wsdl_client.soapproxy.config.debug = 0
        wsdl_client.soapproxy.config.dict_encoding = "UTF-8"
        try:
            result = wsdl_client.cancelarAsincrono(*params)
        except (
            HTTPError, UnknownTypeError, UnderflowError, RecursionError
        ) as e:
            logger.info(e)
        status_cancel = result["status"] or ""
        mensaje_cancel = result["mensaje"] or ""
        cancel_answer = PacCancelAnswer(cfdi.uuid, status_cancel, status_cancel)
        # Something went wrong and raise exception
        if cancel_answer.estatus in pac_exceptions.cancel_error_list:
            raise pac_exceptions.CancelError(
                code=cancel_answer, message=mensaje_cancel,
            )
        # Cancel process was unsuccessful
        elif cancel_answer.estatus not in ("200", "211"):
            raise Exception(
                " ".join(["Code", status_cancel, mensaje_cancel]),
            )
        # If code result is 200 then all was okay
        return cancel_answer.check_cancel_status()

    def _upload_ws_file(self, fdata=None):
        # pylint: disable=too-many-statements
        """
        @params fdata : File.xml codification in base64
        """
        msg = ""
        cfdi_xml = False
        comprobante = "cfdi:Comprobante"
        cfd_data = base64.decodestring(fdata)
        xml_res_str = xml.dom.minidom.parseString(cfd_data)
        xml_res_addenda = self.add_addenta_xml(xml_res_str, comprobante)
        xml_res_str_addenda = xml_res_addenda.toxml("UTF-8")
        xml_res_str_addenda = xml_res_str_addenda.replace(codecs.BOM_UTF8, "")

        # Get values use to connect with SAT
        user = self.user
        password = self.password
        wsdl_url = self.url_webservice_sign
        namespace = self.namespace_sign
        if "testing" in wsdl_url:
            logger.info("WARNING, SIGNED IN TEST!!!!")
            msg += _("WARNING, SIGNED IN TEST!!!!\n\n")
        try:
            wsdl_client = WSDL.SOAPProxy(wsdl_url, namespace)
        except:  # noqa
            # TODO: catch specific errors
            # Send error to be catch in higher levels
            raise
        cfdi = base64.encodestring(xml_res_str_addenda)
        params = [user, password, cfdi, False]
        wsdl_client.soapproxy.config.dumpSOAPOut = 0
        wsdl_client.soapproxy.config.dumpSOAPIn = 0
        wsdl_client.soapproxy.config.debug = 0
        wsdl_client.soapproxy.config.dict_encoding = "UTF-8"
        resultado = wsdl_client.timbrar(*params)
        mensaje = _(tools.ustr(resultado["mensaje"]))
        resultados_mensaje = (
            resultado["resultados"]["mensaje"] if resultado["resultados"] else ""
        )
        folio_fiscal = (
            resultado["resultados"]["uuid"] if resultado["resultados"] else ""
        )
        codigo_timbrado = resultado["status"] or ""
        codigo_validacion = (
            resultado["resultados"]["status"] if resultado["resultados"] else ""
        )
        if codigo_timbrado in ["311", "312"]:
            raise pac_exceptions.PrimarySectorError(code=resultado["codigo"])
        elif (
            codigo_timbrado == "200"
            and codigo_validacion == "200"
            or codigo_validacion == "307"
        ):
            fecha_timbrado = resultado["resultados"]["fechaTimbrado"] or False
            fecha_timbrado = (
                time.strftime(
                    "%Y-%m-%d %H:%M:%S",
                    time.strptime(fecha_timbrado[:19], "%Y-%m-%dT%H:%M:%S"),
                )
                if fecha_timbrado else False
            )
            cfdi_data = {
                "cfdi_xml": base64.decodestring(
                    resultado["resultados"]["cfdiTimbrado"] or ""
                )  # este se necesita en uno que no es base64
            }
            msg += (
                mensaje
                + "."
                + resultados_mensaje
                + " Folio Fiscal: "
                + folio_fiscal
                + "."
            )
            msg += _(
                "Make Sure to the file really has generated correctly to "
                "the SAT\nhttps://www.consulta.sat.gob.mx/sicofi_web/"
                "moduloECFD_plus/ValidadorCFDI/Validador%20cfdi.html"
            )
            if cfdi_data.get("cfdi_xml", False):
                url_pac = (
                    "</'{comprobante}'>"
                    "<!--Para validar el XML CFDI puede descargar el certificado del "
                    "PAC desde la siguiente liga: "
                    "https://solucionfactible.com/cfdi/00001000000102699425.zip"
                    "-->".format(comprobante=comprobante)
                )
                cfdi_data["cfdi_xml"] = cfdi_data["cfdi_xml"].replace(
                    '</"%s">' % (comprobante), url_pac
                )
                f = base64.encodestring(cfdi_data["cfdi_xml"] or "")
                cfdi_xml = cfdi_data.pop("cfdi_xml")
            if cfdi_xml:
                cfdi_data["cfdi_xml"] = cfdi_xml
            else:
                msg += _("Can't extract the file XML of PAC")
            return {"file": f, "msg": msg, "cfdi_xml": cfdi_xml.decode("utf-8")}
        elif codigo_timbrado in pac_exceptions.validation_error_list:
            # El CFDI no ha sido timbrado
            raise pac_exceptions.ValidationError(
                code=codigo_timbrado, message=" ".join([mensaje, resultados_mensaje])
            )
        else:
            raise Exception(
                " ".join(["Code", codigo_timbrado, mensaje, resultados_mensaje])
            )

    @staticmethod
    def add_addenta_xml(xml_res_str=None, comprobante=None):
        """
        @params xml_res_str: File XML
        @params comprobante: Name to the Node that contain the XML
        """

        def add_node(
            node_name, attrs, parent_node, minidom_xml_obj, attrs_types, order=False
        ):  # noqa
            if not order:
                order = attrs
            new_node = minidom_xml_obj.createElement(node_name)
            for key in order:
                if attrs_types[key] == "attribute":
                    new_node.setAttribute(key, attrs[key])
                elif attrs_types[key] == "textNode":
                    key_node = minidom_xml_obj.createElement(key)
                    text_node = minidom_xml_obj.createTextNode(attrs[key])
                    key_node.appendChild(text_node)
                    new_node.appendChild(key_node)
                elif attrs_types[key] == "att_text":
                    new_node.setAttribute("name", key)
                    text_node = minidom_xml_obj.createTextNode(attrs[key])
                    new_node.appendChild(text_node)
            parent_node.appendChild(new_node)
            return new_node

        if xml_res_str:
            node_addenda = xml_res_str.getElementsByTagName("cfdi:Addenda")
            # pylint: disable=len-as-condition
            if len(node_addenda) == 0:
                node_comprobante = xml_res_str.getElementsByTagName(comprobante)[0]
                node_addenda = add_node(
                    "cfdi:Addenda", {}, node_comprobante, xml_res_str, attrs_types={}
                )
                node_partner_attrs = {
                    "xmlns:sf": "http://timbrado.solucionfactible.com/partners",  # noqa
                    "xsi:schemaLocation": '"http://timbrado.solucionfactible.com/partners https://solucionfactible.com/timbrado/partners/partners.xsd',  # noqa
                    "id": "150731",
                }
                node_partner_attrs_types = {
                    "xmlns:sf": "attribute",
                    "xsi:schemaLocation": "attribute",
                    "id": "attribute",
                }
                add_node(
                    "sf:Partner",
                    node_partner_attrs,
                    node_addenda,
                    xml_res_str,
                    attrs_types=node_partner_attrs_types,
                )
            else:
                node_partner_attrs = {
                    "xmlns:sf": "http://timbrado.solucionfactible.com/partners",  # noqa
                    "xsi:schemaLocation": "http://timbrado.solucionfactible.com/partners https://solucionfactible.com/timbrado/partners/partners.xsd",  # noqa
                    "id": "150731",
                }
                node_partner_attrs_types = {
                    "xmlns:sf": "attribute",
                    "xsi:schemaLocation": "attribute",
                    "id": "attribute",
                }
                add_node(
                    "sf:Partner",
                    node_partner_attrs,
                    node_addenda,
                    xml_res_str,
                    attrs_types=node_partner_attrs_types,
                )
        return xml_res_str

    def sf_consult(self, cfdi):
        """ Consul the cfdi status

        :param cfdi: CFDI to check
        """
        method = "consultar"
        try:
            data = sf_envelopes._SOAPENV_CONSULTAR % (
                self.user,
                self.password,
                cfdi.uuid.encode("utf-8"),
            )
        except Exception as e:
            e.message = "Error al crear ENVELOPE"
            logger.info(e)
            raise e
        else:
            logger.debug(data)
        headers = self.create_header(method)
        s = Session()
        req = Request("POST", self.url_webservice_cancel, data=data, headers=headers)
        prepped = req.prepare()
        try:
            # Actually consume service
            response = s.send(prepped, timeout=TIMEOUT)
            res = xmltodict.parse(response.text)
            logger.debug(res)
        except Exception as e:
            # Error al establecer comunicación con el PAC
            logger.debug(e)
            raise e
        # Procesar los resultados obtenidos del PAC
        try:
            resultado = res["soapenv:Envelope"]["soapenv:Body"][
                "ns:getStatusCancelacionAsincronaResponse"
            ][
                "ns:return"
            ]
        except KeyError as e:
            resultado = res["soapenv:Envelope"]["soapenv:Body"][
                "ns2:consultarCfdiResponse"
            ][
                "ns2:consultarCfdiResponse"
            ]
            logger.debug(e)
        except Exception as e:
            logger.debug(e)
            raise UserError(_("Unknow answer.\n%s") % resultado)
        status_cancelacion = resultado["ax27:status"]
        if status_cancelacion in ("211", "213", "204"):
            estado_cfdi = "Vigente"
        elif status_cancelacion in "200":
            estado_cfdi = "Cancelado"
        elif status_cancelacion in "702":
            estado_cfdi = "No encontrado"
        consult_answer = PacConsultAnswer(cfdi.uuid, estado_cfdi, status_cancelacion)
        return consult_answer.check_cancel_status()
